"""
This module contains components that are common to several demo layouts.

"""
import dash_core_components as dcc
import dash_html_components as html
from plotly.subplots import make_subplots
import numpy as np

colors = {
    'background': 'SteelBlue',
    'text': 'White',
    'controls': '#DDDDDD',
    'buttons': 'Orange'
}



def app_parameter(label, appid, inputtype, value, size='20'):
    """
    A label and a Input control, as elements of a table row
    :param label: str, label to show in app layout
    :param appid:  str, label to identify user input within the app
    :param inputtype: str, type of input, either "text" or "number"
    :param value: initial value for this parameter
    :param size: str, size of input field
    :return:
    tuple, a label and a Input control, as html.Th elements to be included in a html.Tr table row
    """
    return html.Th(label), html.Th(dcc.Input(id=appid, type=inputtype, value=value, size=size))


def app_parameter_row(label, appid, inputtype, value, size='20'):
    """
    Table row with a label and a Input control
    :param label: str, label to show in app layout
    :param appid:  str, label to identify user input within the app
    :param inputtype: str, type of input, either "text" or "number"
    :param value: initial value for this parameter
    :param size: str, size of input field
    :return:
    a html.TR table row, with a label and a Input control
    """
    return html.Tr([
        *app_parameter(label, appid, inputtype, value, size)],
        style={'border':'3px dashed SteelBlue'}
    )


def app_choose_parameter(label, appid, list_of_options, value):
    """
    A label and a dropdown menu
    :param label: str, label to show in app layout
    :param appid: str, label to identify user input within the app
    :param list_of_options: list of str, available options to choose from
    :param value: initial value for this parameter
    :return:
    a html.TR table row, with a label and a Dropdown control
    """
    options = [{'label': x, 'value': x} for x in list_of_options]
    return html.Tr([
        html.Th(label),
        html.Th(dcc.Dropdown(id=appid, options=options, value=value), style={'width': '100%'})],
                   style={'border':'3px dashed SteelBlue'}
    )


def app_model_parameter(label, base_value, scen_value):
    """
    A label and two Input controls
    :param label: str, label to show in app layout
    :param base_value: initial baseline value for this parameter
    :param scen_value: initial alternative value for this parameter
    :return:
    a html.TR table row, with a label, an Input control for baseline and a Input control for alternative
    """
    txt = html.Th(label)
    param1 = html.Th(dcc.Input(id='base_' + label, type='text', value=base_value, size='10'))
    param2 = html.Th(dcc.Input(id='scen_' + label, type='text', value=scen_value, size='10'))
    return html.Tr(
        [txt, param1, param2],
        style={'border':'3px dashed SteelBlue'}
    )


def app_one_option(label, appid, list_of_labels, list_of_values, value):
    """
    A label and single choice options, inline
    :param label: str, label to show in app layout
    :param appid: str, label to identify user input within the app
    :param list_of_labels: list of str, available options to show in app
    :param list_of_values: list of str, available options to choose from
    :param value: initial value for this parameter
    :return:
    a html.TR table row, with a label, an RadioItems control
    """

    options = [{'label': x, 'value': y} for x, y in zip(list_of_labels, list_of_values)]
    return html.Tr([
        html.Th(label),
        html.Th(dcc.RadioItems(
            id=appid,
            options=options,
            value=value,
            labelStyle={'display': 'inline-block'}),
            style={'width': '100%'})],
        style={'border':'3px dashed SteelBlue'}
    )


def app_table_headers(labels):
    """
    Headers for table of Inputs
    :param labels: list of str, headers to show on top of inputs
    :return:
    a html.TR table row, with a list of headers
    """
    return html.Tr(
        [html.Th(label) for label in labels],
        style={'border':'3px dashed SteelBlue'}
    )


editable_cell_format = {
    'backgroundColor': 'rgb(75, 75, 75)',
    'color': 'white',
    'fontWeight': 'bold',
    'fontSize': 18
}

header_cell_format = {
    'backgroundColor': 'rgb(50, 50, 50)',
    'color': 'white',
    'fontWeight': 'bold',
    'fontSize': 18
}

def add_plot(fig, existing, **kwargs):
    """
    Add a plotly figure as traces to another figure

    :param fig: a plotly.Figure obtained with make_plots, where to insert next figure
    :param existing: a plotly.Figure to be added as subplot to the figure above
    :param kwargs: parameters passed to method fig.add_trace
    :return: None, updates fig in place
    """
    for ff in existing.select_traces():
        fig.add_trace(ff, **kwargs)


def merge_plots(figures, cols=1, shared_xaxes=True, shared_yaxes=True, subplot_titles=None, title_text=None):
    """
    Merges several plotly.Figures into a single figure (as subplots)

    :param figures: the plotly.Figures to be merged together
    :param cols: number of columns for resulting subplots
    :param shared_xaxes: True if every column must have same x scale
    :param shared_yaxes: True if every row must have same y scale
    :param subplot_titles: a collection of strings to be used as titles for the subplots
    :param title_text: a string to be used as title to the merged figure
    :return:
        a plotly.Figure
    """
    rows = int(np.ceil(len(figures)/cols))
    bigfig = make_subplots(rows, cols, subplot_titles=subplot_titles,shared_xaxes=shared_xaxes, shared_yaxes=shared_yaxes)
    for n, fig in enumerate(figures):
        add_plot(bigfig, fig, row=n // cols + 1, col=n % cols + 1)
    bigfig.update_layout(title_text=title_text, showlegend=False)
    return bigfig

def make_top_banner(title):
    logo = html.Div(
        style={'textAlign': 'center', 'color': colors['controls'], 'width': '5%', 'display': 'inline-block'},
        children=[
            html.Img(src='https://randall-romero.github.io/econometria/_static/r2-logo.png',
                     width='100%',
                     style={'margin-left': '0px',
                            'margin-top': '2px',
                            'margin-bottom': '2px'
                            })])
    header = html.Div(
        style={'textAlign': 'center', 'color': colors['controls'], 'width': '93%', 'display': 'inline-block'},
        children=[
            html.H2(id='title',
                    children=title,
                    style={'textAlign': 'center', 'color': colors['text']})])

    return logo, header

def make_bottom_banner(name_of_demo, width=25):
    footer = html.Div(
        style={'textAlign': 'center', 'color': colors['controls'], 'width': '100%', 'display': 'inline-block'},
        children=[
            html.Div(style={'textAlign': 'center', 'color': colors['controls'], 'width': f'{width}%', 'display': 'inline-block'},
                     children=[html.P(name_of_demo,
                                      style={'textAlign': 'center', 'color': colors['text'] ,'marginTop' :10})]),
            html.Div(style={'textAlign': 'left', 'color': colors['controls'], 'width': f'{100-width}%', 'display': 'inline-block'},
                     children=[html.A(children='Randall Romero Aguilar',
                                      href='http://randall-romero.com',
                                      style={'textAlign': 'left', 'color': colors['text'] ,'marginBottom' :20})])
        ]
    )

    return footer



"""
macrodemos : A Python module to teach macroeconomics and macroeconometrics

This package contains Dash apps to teach and learn about macroeconomics and macroeconometrics.

So far, the available demos are:

    ARMA_demo():  to study AutoRegressive Moving Average processes.
    filters_demo(): to compare several filtering methods.
    Markov_demo(): to study Markov chains.
    Solow_demo():  to study the Solow growth model.

Randall Romero Aguilar, 2016-2021
"""

#import .constants
#import .common_components
from .captura import read_IFS
from .demo_ARMA import ARMA
from .demo_Solow import Solow
from .demo_Markov import Markov
from .demo_filters import filters
# from .demo_BoxJenkins import BoxJenkins # still work in progress





from jupyter_dash import JupyterDash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
import plotly.express as px
from dash_table import DataTable


import webbrowser

from macrodemos.common_components import app_model_parameter, app_table_headers, editable_cell_format, header_cell_format, \
    make_bottom_banner, colors, make_top_banner


import numpy as np
import pandas as pd


# Esta parte controla asuntos de estética de la página
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']


class SolowSwan:
    names = dict(
        A='total productivity of factors',
        s='marginal rate of savings',
        α='marginal product of capital',
        δ='depreciation rate',
        n='population growth rate',
        k='Capital Stock (per capita)',
        y='Output per capita',
        sy='Savings per capita',
        c='Consumption per capita',
        i='Investment per capita',
        Δk='Change in capital stock',
        gy='Output growth rate'
    )

    math_form = dict(
        A=r'$A$',
        s=r'$s$',
        α=r'$\alpha$',
        δ=r'$\delta$',
        n=r'$n$',
        k=r'$k_t$',
        y=r'$y_t$',
        sy=r'$sy_t$',
        c=r'$c_t$',
        i=r'$i_t$',
        Δk=r'$\Delta k_t$',
        gy=r'${g_y}_t$'
    )
    endogenous = ['k', 'y', 'sy', 'c', 'Δk', 'gy']

    def __init__(self, A, s, α, δ, n):
        """

        :param A: float, total productivity of factors
        :param s: float, marginal rate of savings
        :param α: float, marginal product of capital
        :param δ: float, depreciation rate
        :param n: float, population growth rate
        """
        self.A = A
        self.s = s
        self.α = α
        self.δ = δ
        self.n = n
        self.data = None
        self.steady_state = dict()
        self.compute_steady_state()

    @property
    def parameters(self):
        return self.A, self.s, self.α, self.δ, self.n

    def f(self, k):
        return self.A * k ** self.α

    def compute_steady_state(self):
        A, s, α, δ, n = self.parameters

        k = (s * A / (n + δ)) ** (1 / (1 - α))
        y = self.f(k)
        i = (n + δ) * k
        c = y - i

        for name, value in zip('ykic', [y, k, i, c]):
            self.steady_state[name] = value

        self.steady_state['Δk'] = 0.0
        self.steady_state['gy'] = 0.0
        self.steady_state['sy'] = s * y

    def compute_ksteady(self, s):
        A, _, α, δ, n = self.parameters
        k = (s * A / (n + δ)) ** (1 / (1 - α))
        y = self.f(k)
        i = (n + δ) * k
        c = y - i
        return pd.DataFrame({'k': k, 'y': y, 'i': i, 'c': c}, index=s)

    def simulate(self, T, K0=None):
        A, s, α, δ, n = self.parameters

        if K0 is None:
            K0 = self.steady_state['k']

        Y0 = self.f(K0)
        S0 = s * Y0

        K, S, Y = np.zeros([3, T + 1], dtype=float)
        K[0], S[0], Y[0] = K0, S0, Y0

        for t in range(T):
            K[t + 1] = ((1 - δ) * K[t] + S[t]) / (1 + n)
            Y[t + 1] = A * K[t + 1] ** α
            S[t + 1] = s * Y[t + 1]

        datos = pd.DataFrame({'k': K, 'y': Y, 'sy': S})
        datos['c'] = Y - S
        datos['Δk'] = (S - (n + δ) * K) / (1 + n)
        datos.loc[0, 'Δk'] = (S[0] - (self.n + self.δ) * K[0]) / (1 + self.n)
        datos['gy'] = datos['Δk'] * datos['k']

        self.data = datos

    def plot_compare_paths(self, other, ser):
        df = pd.concat([self.data[ser], other.data[ser]],
                       keys=['Baseline', 'Alternative'],
                       axis=1)
        fig = px.line(df,
                      y=df.columns,
                      title=self.names[ser],
                      template='simple_white'
                      )
        fig.update_layout(legend_orientation='h',
                          xaxis_title='',
                          yaxis_title='')

        fig.add_annotation(x=0.1,
                           y=0.9,
                           xref='paper',
                           yref='paper',
                           showarrow=False,
                           text=self.math_form[ser],
                           font=dict(size=30)
                           )
        return fig

    def plot_golden(self, df, ser):
        fig = px.line(df,
                      y=['Baseline', 'Alternative'],
                      title=self.names[ser],
                      template='simple_white'
                      )
        fig.update_layout(#legend_orientation='h',
                          xaxis_title='Savings rate',
                          yaxis_title='')

        fig.add_annotation(x=0.1,
                           y=0.9,
                           xref='paper',
                           yref='paper',
                           showarrow=False,
                           text=self.math_form[ser],
                           font=dict(size=30)
                           )

        return fig

    def plot_compare_golden_rule(self, other):
        s = np.linspace(0, 1, 201)
        df = pd.concat([self.compute_ksteady(s), other.compute_ksteady(s)],
                       axis=1,
                       keys=['Baseline', 'Alternative']).swaplevel(axis=1)

        return [self.plot_golden(df[ser], ser) for ser in 'ciky']

    def plot_compare_steady_state(self, other):

        A0, s0, α0, δ0, n0 = self.parameters
        A1, s1, α1, δ1, n1 = other.parameters

        kmax = max(self.steady_state['k'], other.steady_state['k'])

        k = np.linspace(0, 1.2 * kmax, 201)

        df = pd.DataFrame({
            '$f(k)_0$': self.f(k),
            '$sf(k)_0$': s0 * self.f(k),
            '$(n+δ)k_0$': (n0 + δ0) * k,
            '$f(k)_1$': other.f(k),
            '$sf(k)_1$': s1 * other.f(k),
            '$(n+δ)k_1$': (n1 + δ1) * k},
            index=k)

        fig = px.line(df,
                      y=df.columns,
                      title='Determining the steady-state level of capital',
                      template='simple_white'
                      )
        fig.update_layout(height=600,
                          legend_orientation='h',
                          xaxis_title='$k_t$',
                          yaxis_title='$f(k_t),\quad sf(k_t)$')

        fig.add_annotation(x=self.steady_state['k'],
                           y=self.steady_state['sy'],
                           showarrow=False,
                           text='$k^*_0$',
                           font=dict(size=20)
                           )

        fig.add_annotation(x=other.steady_state['k'],
                           y=other.steady_state['sy'],
                           showarrow=False,
                           text='$k^*_1$',
                           font=dict(size=20)
                           )

        return fig

#=======================================================================================================================
#
#  APP STARTS HERE
#
#_______________________________________________________________________________________________________________________
mathjax = ["https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.4/MathJax.js?config=TeX-MML-AM_CHTML"]  # to display math
omega = np.linspace(0, np.pi, 121)  # frequencies for spectral plot


app = JupyterDash(__name__,external_stylesheets=external_stylesheets, external_scripts=mathjax)

#======DESIGN THE APP===============

app.layout = html.Div(children=[
    # TOP BANNER-------
    *make_top_banner('The Solow-Swan Model'),
    # ==================================================================================================================
    #            INICIO DE COLUMNA DE PARÁMETROS
    # ------------------------------------------------------------------------------------------------------------------
    html.Div(
        style={'textAlign': 'center', 'color': colors['controls'], 'width': '25%', 'display': 'inline-block'},
        children=[
            html.H4("Parameters"),
            html.Table(children=[
                app_table_headers(['  ', 'Baseline', 'Alternative']),
                app_model_parameter('α', 0.35, 0.35),
                app_model_parameter('δ', 0.06, 0.06),
                app_model_parameter('n', 0.02, 0.02)]
            ),
            html.Hr(),
            html.H4("Exogenous variables"),
            html.Table(children=[
                app_table_headers(['  ', 'Baseline', 'Alternative']),
                app_model_parameter('A', 1.0, 1.0),
                app_model_parameter('s', 0.2, 0.2)]
                ),
            html.Hr(),
            html.H4("Steady state"),
            html.Div(id='output-data-upload', children=[
            DataTable(
                id='table-transition-matrix',
                columns=([{'id': p, 'name': p} for p in ['Variable', 'Baseline', 'Alternative', '% change']]),
                editable=False,
                #style_table={'width': '75%'},
                style_cell=editable_cell_format,
                style_header=header_cell_format
            )
        ],
                 ),
        html.Hr(),
        html.H4("Figure parameter"),
        html.Table(children=[
            html.Tr([html.Th('Number of periods'), dcc.Input(id='horizon', type='text', value=60, size='10')])]
        ),
        html.Button('PLOT', id='execute',style={'textAlign': 'center', 'backgroundColor': colors['buttons']}),
        html.P('Based on chapter 9 of Bongers, Gómez and Torres (2019) Introducción a la Macroeconomía Computacional. Vernon Press.',
               style={'textAlign': 'left', 'color': colors['text'],'marginTop':20}),
    ],
        ),
    # ==================================================================================================================
    #            INICIO DE COLUMNA DE RESULTADOS
    # ------------------------------------------------------------------------------------------------------------------
    html.Div(style={'width': '75%', 'float': 'right', 'display': 'inline-block'},
             children=[dcc.Tabs([
                 # Panel 1: Simulation==================================================================================
                 dcc.Tab(label="Simulations",
                         children=[
                             # --PLOT 1a:  capital stock----------------------------------------------------------------
                             dcc.Graph(id='path-capital',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 1b:  output per capita ------------------------------------------------------------
                             dcc.Graph(id='path-output',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 1c: savings------------------------------------------------------------------------
                             dcc.Graph(id='path-savings',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             # --PLOT 2a:  consumption------------------------------------------------------------------
                             dcc.Graph(id='path-consumption',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 2b: change in capital--------------------------------------------------------------
                             dcc.Graph(id='path-delta-capital',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 2c:  Growth rate ------------------------------------------------------------------
                             dcc.Graph(id='path-growth',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'})
                         ]),
                 # Panel 2: Equilibrium=================================================================================
                 dcc.Tab(label="Equilibrium",
                         children=[
                             # --PLOT 1a:  Equilibrium----------------------------------------------------------------
                             dcc.Graph(id='model-equilibrium',
                                       style={'width': '98%', 'float': 'left', 'display': 'inline-block'}),
                         ]
                         ),
                 # Panel 3: Golden Rule=================================================================================
                 dcc.Tab(label="Golden Rule",
                         children=[
                             # --PLOT 1a:  capital stock----------------------------------------------------------------
                             dcc.Graph(id='golden-capital',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 1b:  output per capita ------------------------------------------------------------
                             dcc.Graph(id='golden-output',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             #--PLOT 1c: savings------------------------------------------------------------------------
                             dcc.Graph(id='golden-savings',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                             # --PLOT 2a:  consumption------------------------------------------------------------------
                             dcc.Graph(id='golden-consumption',
                                       style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                         ]
                         ),
             ]), # end of Tabs
             ]),
    make_bottom_banner("macrodemos.Solow( )")
],
    style={'backgroundColor': colors['background']})


@app.callback(
    [Output('path-capital', 'figure'),
     Output('path-output', 'figure'),
     Output('path-savings', 'figure'),
     Output('path-consumption', 'figure'),
     Output('path-delta-capital', 'figure'),
     Output('path-growth', 'figure'),
     Output('table-transition-matrix', 'data'),
     Output('golden-capital', 'figure'),
     Output('golden-output', 'figure'),
     Output('golden-savings', 'figure'),
     Output('golden-consumption', 'figure'),
     Output('model-equilibrium', 'figure'),
     ],
    [Input('execute', 'n_clicks')],
    [State('horizon','value'),
     State('base_α', 'value'),
     State('base_δ', 'value'),
     State('base_n', 'value'),
     State('base_A', 'value'),
     State('base_s', 'value'),
     State('scen_α', 'value'),
     State('scen_δ', 'value'),
     State('scen_n', 'value'),
     State('scen_A', 'value'),
     State('scen_s', 'value')])
def update_Solow_parameters(n_clicks,T, α, δ, n, A, s, α1, δ1, n1, A1, s1):
    # baseline scenario
    baseline = SolowSwan(*[float(xx) for xx in (A, s, α, δ, n)])
    baseline.simulate(T=int(T))

    #alternative scenario
    alternative = SolowSwan(*[float(xx) for xx in (A1, s1, α1, δ1, n1)])
    alternative.simulate(T=int(T), K0=baseline.steady_state['k'])

    # Plots for TAB Simulation
    simulation_plots = [baseline.plot_compare_paths(alternative,ser) for ser in baseline.endogenous]

    # Plots for TAB Golden
    golden_plots = baseline.plot_compare_golden_rule(alternative)

    #Plot for TAB Equilibrium
    equilibrium_plots = [baseline.plot_compare_steady_state(alternative)]

    # Table to compare the steady-states of baseline v. alternative
    df = pd.DataFrame({'Baseline': [v for v in baseline.steady_state.values()],
                      'Alternative': [v for v in alternative.steady_state.values()]},
                      index=baseline.steady_state.keys()
                      )
    df['% change'] = 100 * (df['Alternative'] / df['Baseline'] - 1)
    df = df.round(3).loc[baseline.endogenous]
    df.index.name = 'Variable'
    steady_state_table = [df.reset_index().to_dict('records')]

    return simulation_plots + steady_state_table + golden_plots + equilibrium_plots




def Solow(colab=False):
    if colab:
        app.run_server(mode='external')
    else:
        webbrowser.open('http://127.0.0.1:8050/')
        app.run_server(debug=False)


Solow_demo = Solow # kept for backward compatibility

if __name__ == '__main__':
    Solow()


# TODO : plot forecast
# TODO: report estimation outputs


import base64
import datetime
import io


from jupyter_dash import JupyterDash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash_table import DataTable
import plotly.express as px
import plotly.graph_objects as go

import webbrowser

import numpy as np
import pandas as pd
pd.options.plotting.backend = "plotly"


from statsmodels.tsa.arima.model import ARIMA as tsaARMA # PARA ESTIMAR MODELO
from statsmodels.tsa.stattools import acf, pacf # autocorrelation and partial autocorrelation

from macrodemos.common_components import app_parameter, app_parameter_row, editable_cell_format, \
    colors, make_bottom_banner, make_top_banner, app_choose_parameter, app_one_option, app_model_parameter, app_table_headers




import warnings


omega = np.linspace(0, np.pi, 121)  # frequencies for spectral plot

def add_polar_scatter(fig, roots, name, ms=16):
    """
    Adds a polar scatter to existing figure

    To help making the plots for the roots of the ARIMA polynomial

    :param fig: existing plotly.Figure
    :param roots: roots of the characteristic polynomial
    :param name: name for the series (to use in legend)
    :param ms: marker size
    :return:
    float, maximum absolute value of roots.
    Updates fig in place
    """
    radium = np.abs(roots)
    angles = np.angle(roots, True)
    fig.add_trace(go.Scatterpolar(
        r=radium,
        theta=angles,
        name = name,
        mode = 'markers',
        marker_size=ms
    ))
    return max(radium)

def compute_spectral(yvalues):
    yy = yvalues - yvalues.mean()
    v = (yy ** 2).mean()  # estimated variance
    T = yy.size

    r = int(np.sqrt(T))
    gamma = acf(yy, adjusted=False, nlags=r, fft=False)
    k = np.arange(r + 1)
    g = 1 - k / r # Bartlett window
    sw = ((np.cos(np.outer(omega, k)) * (g * gamma)).sum(axis=1) * 2 - 1)
    sw *= v / (2 * np.pi)  # rescale
    return sw


def parse_contents(contents, filename, date):
    content_type, content_string = contents.split(',')

    decoded = base64.b64decode(content_string)
    try:
        if 'csv' in filename[-4:]:
            # Assume that the user uploaded a CSV file
            df = pd.read_csv(
                io.StringIO(decoded.decode('utf-8')), parse_dates=[0])
        elif 'xls' in filename[-4:]:
            # Assume that the user uploaded an Excel file
            df = pd.read_excel(io.BytesIO(decoded), parse_dates=[0])
    except Exception as e:
        print(e)
        return html.Div([
            'There was an error processing this file.'
        ])

    df.set_index(df.columns[0], inplace=True)

    return df


# =======================================================================================================================
#
#  ARMA_class CLASS
#
# _______________________________________________________________________________________________________________________


class ARIMA:
    def __init__(self, df):
        self.data = df
        self.y = None
        self.current_variable_name = None
        self.estimates = [{}, {}]
        self.estimates[0]['residuals'] = np.zeros(5)
        self.estimates[1]['residuals'] = np.zeros(5)
        self.estimated = [False, False]
        self.estimation_order = [dict(p=0, q=0), dict(p=0, q=0)]

    def transform_series(self, varname, take_log, diff):
        serie = self.data[varname]
        name = varname

        if take_log == 'yes':
            serie = np.log(serie)
            name = f'log({name})'

        if diff == '1':
            serie = serie.diff()
            name = f'Δ{name}'

        if diff == '2':
            serie = serie.diff().diff()
            name = f'Δ2{name}'

        self.y = serie.dropna()
        self.current_variable_name = name

        # ACTUAL DATA PLOTS-----------------------------------------

    def plot_actual_data(self):
        return self.y.plot(title=self.current_variable_name)
        # return px.line(y=self.y.values,
        #               title=self.current_variable_name,
        #               template='simple_white'
        #              )

    def plot_correlogram(self, lags):
        return px.bar(y=acf(self.y.values, adjusted=True, nlags=lags, fft=False),
                      title='Autocorrelations',
                      labels=dict(x='lag', y='coefficient'),
                      # color_discrete_sequence=px.colors.qualitative.Dark24,
                      template='simple_white'
                      )

    def plot_partial_correlogram(self, lags):
        return px.bar(y=pacf(self.y.values, nlags=lags),
                      title='Partial Autocorrelations',
                      labels=dict(x='lag', y='coefficient'),
                      # color_discrete_sequence=px.colors.qualitative.Dark24,
                      template='simple_white'
                      )

    def plot_spectral(self):

        return px.area(
            x=omega,
            y=compute_spectral(self.y.values),
            title='Spectral Density',
            labels=dict(x=r'$\omega$', y=r'$s(\omega)$'),
            # color_discrete_sequence=px.colors.qualitative.Dark24,
            template='simple_white'
        )

    # ESTIMATION OF MODELS--------------------------------------------------
    def estimate(self, p, q, modelo):
        try:
            print(f'Trying to estimate the model ARMA({p}, {q}) ')
            # res = tsaARMA(self.y, order=[p, 0, q], freq=self.data.index.freqstr).fit()
            res = tsaARMA(self.y, order=[p, 0, q]).fit()
            self.estimates[modelo] = {
                'c': res.params[0],
                'phi': res.arparams if p else None,
                'theta': res.maparams if q else None,
                'fitted': pd.Series(res.fittedvalues),
                'residuals': pd.Series(res.resid),
                'arroots': np.array([1 / x for x in res.arroots]) if p else None,
                'maroots': np.array([1 / x for x in res.maroots]) if q else None
            }
            self.estimated[modelo] = True
            self.estimation_order[modelo] = dict(p=p, q=q)

            res1 = res.summary().tables[1]
            result = pd.DataFrame(res1.data[1:], columns=res1.data[0])
            result = result.iloc[:, [0, 1, 4]]
            result.rename(columns={'': 'param'}, inplace=True)

            new_estimate = DataTable(
                data=result.to_dict('records'),
                columns=[{'name': i, 'id': i} for i in result.columns])
            self.estimated_table = new_estimate.data
            print(f'Model {modelo} succesfully estimated!')

        except:
            self.estimated[modelo] = False
            print('The model could not be estimated')
            # self.estimates['repr'] = 'The model could not be estimated'
            self.estimated_table = []

            # RESIDUAL PLOTS-------------------------------------------

    def plot_residual(self):
        residuos = pd.DataFrame(
            {'Model 1': self.estimates[0]['residuals'],
             'Model 2': self.estimates[1]['residuals']},
            index=self.data.index
        )
        return residuos.plot(title='Model Fit Residuals')

    def plot_residual_correlogram(self, nlags):
        df = pd.DataFrame(
            {'Model 1': acf(self.estimates[0]['residuals'].values, adjusted=True, nlags=nlags, fft=False),
             'Model 2': acf(self.estimates[1]['residuals'].values, adjusted=True, nlags=nlags, fft=False)}
        )

        fig = px.bar(df.iloc[1:],
                     barmode='group',
                     title='Autocorrelations of Model Residuals',
                     template='simple_white')

        fig.update_layout(
            legend_orientation='h',
            xaxis_title='Lag',
            yaxis_title=r'$\rho$'
        )
        return fig

    def plot_residual_partial_correlogram(self, nlags):
        df = pd.DataFrame(
            {'Model 1': pacf(self.estimates[0]['residuals'].values, nlags=nlags),
             'Model 2': pacf(self.estimates[1]['residuals'].values, nlags=nlags)}
        )

        fig = px.bar(df.iloc[1:],
                     barmode='group',
                     title='Partial Autocorrelations of Model Residuals',
                     template='simple_white')

        fig.update_layout(
            legend_orientation='h',
            xaxis_title='Lag',
            yaxis_title=r'$\rho$'
        )
        return fig

    def plot_residual_spectral(self):

        df = pd.DataFrame(
            {'Model 1': compute_spectral(self.estimates[0]['residuals'].dropna().values),
             'Model 2': compute_spectral(self.estimates[1]['residuals'].dropna().values),
             'Frequency': omega}
        )

        # sw0 = compute_spectral(self.estimates[0]['residuals'].dropna().values)
        # sw1 = compute_spectral(self.estimates[1]['residuals'].dropna().values)

        fig = px.area(
            df,
            x='Frequency',
            y=['Model 1'],
            title='Spectral Density',
            labels=dict(x=r'$\omega$', y=r'$s(\omega)$'),
            # color_discrete_sequence=px.colors.qualitative.Dark24,
            template='simple_white'
        )

        fig.add_trace(go.Scatter(
            x=df['Frequency'],
            y=df['Model 2'],
            fill='tozeroy',
            mode='lines',
            name='Model 2'
        )
        )
        return fig

    def plot_ar_roots(self):
        fig = go.Figure()
        rmax0 = add_polar_scatter(fig, self.estimates[0]['arroots'] if self.estimation_order[0]['p'] else [0],
                                  'Model 1')
        rmax1 = add_polar_scatter(fig, self.estimates[1]['arroots'] if self.estimation_order[1]['p'] else [0],
                                  'Model 2')

        maxr = max([1, rmax0, rmax1])
        fig.update_layout(
            title='AR inverse roots',
            polar={'angularaxis': {'thetaunit': 'radians', 'dtick': np.pi / 4},
                   'radialaxis': {'tickvals': [0.0, 1.0], 'range': [0, maxr]}}
        )
        return fig

    def plot_ma_roots(self):
        fig = go.Figure()
        rmax0 = add_polar_scatter(fig, self.estimates[0]['maroots'] if self.estimation_order[0]['q'] else [0],
                                  'Model 1')
        rmax1 = add_polar_scatter(fig, self.estimates[1]['maroots'] if self.estimation_order[1]['q'] else [0],
                                  'Model 2')

        maxr = max([1, rmax0, rmax1])
        fig.update_layout(
            title='MA inverse roots',
            polar={'angularaxis': {'thetaunit': 'radians', 'dtick': np.pi / 4},
                   'radialaxis': {'tickvals': [0.0, 1.0], 'range': [0, maxr]}}
        )
        return fig


APPDATA = ARIMA(None)

# Esta parte controla asuntos de estética de la página
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

# =======================================================================================================================
#
#  APP STARTS HERE
#
# _______________________________________________________________________________________________________________________
mathjax = ["https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.4/MathJax.js?config=TeX-MML-AM_CHTML"]  # to display math



app = JupyterDash(__name__, external_stylesheets=external_stylesheets, external_scripts=mathjax,
                  suppress_callback_exceptions=True)

# ======DESIGN THE APP===============

app.layout = html.Div(children=[
    *make_top_banner("Box Jenkins methodology"),
    html.Div(children=[
        html.H4("Getting data ready"),
        dcc.Upload(
            id='upload-data',
            children=html.Div([
                'Drag and Drop or ',
                html.A('select a CSV or Excel file')
            ]),
            style={
                'width': '100%',
                'height': '60px',
                'lineHeight': '60px',
                'borderWidth': '1px',
                'borderStyle': 'dashed',
                'borderRadius': '5px',
                'textAlign': 'center',
                'margin': '10px'
            },
            multiple=False  # Allow multiple files to be uploaded
        ),
        html.Table(id='select-variable', style={'width': '100%'},
                   # children=[app_choose_parameter('Variable', 'variable', UPLOADED_DATA.columns.tolist(), UPLOADED_DATA.columns[0])]
                   ),
        html.Table(children=[
        ]
        ),
        html.Hr(),
        html.H4("Model specifications"),
        html.Table(children=[
            app_table_headers(['   ', 'Model 1', 'Model 2']),
            app_model_parameter('p', 1, 0),
            app_model_parameter('q', 0, 1),
        ],
        ),
        html.Hr(),
        html.H4("Figure Parameters"),
        html.Table(children=[
            app_parameter_row('AC and PAC lags', 'lags', 'text', '12', '8'),
            app_parameter_row('IRF horizon', 'horizon', 'text', '24', '8')]),
        html.Hr(),
        html.Button('ESTIMATE', id='estimate', style={'textAlign': 'center', 'backgroundColor': colors['buttons']}),
    ],
        style={'textAlign': 'center', 'color': colors['controls'], 'width': '20%', 'display': 'inline-block'}),

    html.Div(style={'width': '80%', 'float': 'right', 'display': 'inline-block'},
             children=[
                 dcc.Tabs([
                     dcc.Tab(label='Data plots',
                             children=[
                                 # --PLOT 1:  SIMULATED TIME SERIES--------------------------------------------------
                                 dcc.Graph(id='plot-data-series',
                                           style={'width': '99%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2a:  AUTOCORRELOGRAM---------------------------------------------------------
                                 dcc.Graph(id='plot-data-acf',
                                           style={'width': '33%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2b:  PARCIAL AUTOCORRELOGRAM-------------------------------------------------
                                 dcc.Graph(id='plot-data-pacf',
                                           style={'width': '33%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2a:  SPECTRAL DENSITY---------------------------------------------------------
                                 dcc.Graph(id='plot-data-spectrum',
                                           style={'width': '33%', 'float': 'left', 'display': 'inline-block'}),
                             ]
                             ),
                     dcc.Tab(label='Estimation plots',
                             children=[
                                 # --PLOT 1:  Residuals--------------------------------------------------
                                 dcc.Graph(id='plot-residual-series',
                                           style={'width': '99%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2a:  AUTOCORRELOGRAM---------------------------------------------------------
                                 dcc.Graph(id='plot-residual-acf',
                                           style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2b:  PARCIAL AUTOCORRELOGRAM-------------------------------------------------
                                 dcc.Graph(id='plot-residual-pacf',
                                           style={'width': '49%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 2a:  SPECTRAL DENSITY---------------------------------------------------------
                                 dcc.Graph(id='plot-residual-spectrum',
                                           style={'width': '39%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 3a:  AR ROOTS-----------------------------------------------------------------
                                 dcc.Graph(id='plot-estimated-ar-roots',
                                           style={'width': '30%', 'float': 'left', 'display': 'inline-block'}),
                                 # --PLOT 3b:  IMPULSE RESPONSE FUNCTION------------------------------------------------
                                 dcc.Graph(id='plot-estimated-ma-roots',
                                           style={'width': '30%', 'float': 'left', 'display': 'inline-block'}),
                             ]
                             ),
                     dcc.Tab(label='Uploaded data',
                             children=html.Div(id='output-data-upload')),
                 ]),
             ]),
    make_bottom_banner("macrodemos.BoxJenkins()", width=20)
],
    style={'backgroundColor': colors['background']})


def parse_contents(contents, filename):
    content_type, content_string = contents.split(',')

    decoded = base64.b64decode(content_string)
    try:
        if 'csv' in filename[-4:]:
            # Assume that the user uploaded a CSV file
            df = pd.read_csv(
                io.StringIO(decoded.decode('utf-8')), parse_dates=[0])
        elif 'xls' in filename[-4:]:
            # Assume that the user uploaded an excel file
            df = pd.read_excel(io.BytesIO(decoded), parse_dates=[0])
    except Exception as e:
        print(e)
        # return html.Div([
        #    'There was an error processing this file.'
        # ])

    df.set_index(df.columns[0], inplace=True)

    return df


@app.callback([Output('select-variable', 'children'),
               Output('output-data-upload', 'children')],
              [Input('upload-data', 'contents')],
              [State('upload-data', 'filename'),
               State('upload-data', 'last_modified')])
def update_variable_selector(contents, filename, date):
    if contents is not None:
        df = parse_contents(contents, filename)
        APPDATA.data = df
        indicators = df.columns
        variable_selector = [
            app_choose_parameter('Variable', 'variable', indicators, indicators[0]),
            app_one_option('Take log?', 'log', ['yes', 'no'], ['yes', 'no'], 'no'),
            app_one_option('Differentiate?', 'I', ['no', 'once', 'twice'], ['0', '1', '2'], '0')
        ]
        data_table = html.Div([
            html.H5(f'Uploaded file: {filename}',
                    style={'color': 'white'}),
            html.H6(f'Last modified: {datetime.datetime.fromtimestamp(date)}',
                    style={'color': 'white'}),

            DataTable(
                data=df.reset_index().to_dict('records'),
                columns=[{'name': i, 'id': i} for i in df.columns])
        ])
        return variable_selector, data_table


@app.callback([Output('plot-data-series', 'figure'),  # fix this
               Output('plot-data-acf', 'figure'),
               Output('plot-data-pacf', 'figure'),
               Output('plot-data-spectrum', 'figure'),
               ],  # fix this
              [Input('variable', 'value'),
               Input('log', 'value'),
               Input('I', 'value')],
              [State('lags', 'value'),
               State('upload-data', 'contents')]
              )
def prepare_data(varname, take_log, diff, nlags, fileuploaded):
    if fileuploaded:
        APPDATA.transform_series(varname, take_log, diff)
        nlags = int(nlags)
        return (APPDATA.plot_actual_data(),
                APPDATA.plot_correlogram(nlags),
                APPDATA.plot_partial_correlogram(nlags),
                APPDATA.plot_spectral()
                )


@app.callback([Output('plot-residual-series', 'figure'),  # fix this
               Output('plot-residual-acf', 'figure'),
               Output('plot-residual-pacf', 'figure'),
               Output('plot-residual-spectrum', 'figure'),
               Output('plot-estimated-ar-roots', 'figure'),
               Output('plot-estimated-ma-roots', 'figure'),
               ],  # fix this
              [Input('estimate', 'n_clicks')],
              [State('base_p', 'value'),
               State('base_q', 'value'),
               State('scen_p', 'value'),
               State('scen_q', 'value'),
               State('lags', 'value')]
              )
def estimate_models(n_clicks, p1, q1, p2, q2, nlags):
    APPDATA.estimate(int(p1), int(q1), 0)
    APPDATA.estimate(int(p2), int(q2), 1)

    nlags = int(nlags)

    return (
        APPDATA.plot_residual(),
        APPDATA.plot_residual_correlogram(nlags),
        APPDATA.plot_residual_partial_correlogram(nlags),
        APPDATA.plot_residual_spectral(),
        APPDATA.plot_ar_roots(),
        APPDATA.plot_ma_roots()
    )


def BoxJenkins(colab=True):
    if colab:
        app.run_server(mode='external')
    else:
        webbrowser.open('http://127.0.0.1:8050/')
        app.run_server(debug=False)

BoxJenkins_demo = BoxJenkins # kept to keep consistency with all names


if __name__ == '__main__':
    BoxJenkins()